// stuff for Winamp interface

#include <windows.h>
#include <windowsx.h>
#include <mmreg.h>
#include <msacm.h>
#include <math.h>
#include <stdio.h>
#include "in2.h"
#include "wamain.h"
#include "dsp.h"
#include "resource.h"

#define BinaryName "in_dsp.dll"
#define AppName "GC DSP Player"

#ifdef _DEBUG

void __cdecl DisplayError (char * Message, ...) {
	char Msg[400];
	va_list ap;

	va_start( ap, Message );
	vsprintf( Msg, Message, ap );
	va_end( ap );
	MessageBox(NULL,Msg,"Error",MB_OK|MB_ICONERROR|MB_SETFOREGROUND);
}
#else
void __cdecl DisplayError (char * Message, ...) {}
#endif

// avoid CRT. Evil. Big. Bloated.
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}

DWORD WINAPI __stdcall DecodeThread(void *b); // the decode thread procedure

#define WM_WA_MPEG_EOF WM_USER+2 // post this to the main window at end of file (after playback as stopped)
char lastfn[MAX_PATH]; // currently playing file (used for getting info on the current file)

char sample_buffer[576*2*2]; // sample buffer

double decode_pos_ms; // current decoding position, in milliseconds
int paused; // are we paused?
int seek_needed; // if != -1, it is the point that the decode thread should seek to, in ms.
int killDecodeThread=0; // the kill switch for the decode thread

double fadecount;

int looptimes = 2;
int masterfadelength = 10;

In_Module mod; // the input module (declared near the bottom of this file)

//HANDLE input_file=INVALID_HANDLE_VALUE; // input file handle
HANDLE thread_handle=INVALID_HANDLE_VALUE;	// the handle to the decode thread

// uses the configuration file plugin.ini in the same dir as the DLL (a la HE)
void GetINIFileName(char * iniFile) {
	if (GetModuleFileName(GetModuleHandle(BinaryName), iniFile, MAX_PATH)) {
		char * lastSlash = strrchr(iniFile, '\\');
				
		*(lastSlash + 1) = 0;
		strcat(iniFile, "plugin.ini");
	}
}

void WriteSettings(void);

void ReadSettings(void) {
	char iniFile[MAX_PATH+1];
	char buf[256];

	GetINIFileName(iniFile);

	GetPrivateProfileString(AppName,"fadelength","10",buf,sizeof(buf),iniFile);
	sscanf(buf,"%i",&masterfadelength);

	GetPrivateProfileString(AppName,"looptimes","2",buf,sizeof(buf),iniFile);
	sscanf(buf,"%i",&looptimes);
}

BOOL CALLBACK configDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	char buf[256];
	char iniFile[MAX_PATH];
	switch (uMsg) {	
	case WM_CLOSE:
		EndDialog(hDlg,TRUE);
		return 0;
	case WM_INITDIALOG:
		GetINIFileName(iniFile);
		GetPrivateProfileString(AppName,"fadelength","10",buf,sizeof(buf),iniFile);
		SetDlgItemText(hDlg,IDC_FADELEN,buf);
		GetPrivateProfileString(AppName,"looptimes","2",buf,sizeof(buf),iniFile);
		SetDlgItemText(hDlg,IDC_LOOPS,buf);
		
		break;
	case WM_COMMAND:
		switch (GET_WM_COMMAND_ID(wParam, lParam)) {
		case IDOK:
			GetINIFileName(iniFile);
			GetDlgItemText(hDlg,IDC_FADELEN,buf,sizeof(buf)-1);
			sscanf(buf,"%i",&masterfadelength);
			WritePrivateProfileString(AppName,"fadelength",buf,iniFile);
			GetDlgItemText(hDlg,IDC_LOOPS,buf,sizeof(buf)-1);
			sscanf(buf,"%i",&looptimes);
			WritePrivateProfileString(AppName,"looptimes",buf,iniFile);			
		case IDCANCEL:
			EndDialog(hDlg,TRUE);
			break;
		}
	default:
		return 0;
	}

	return 1;
}

void config(HWND hwndParent)
{
	DialogBox(mod.hDllInstance, (const char *)IDD_CONFIG, hwndParent, configDlgProc);
}

void about(HWND hwndParent)
{
	MessageBox(hwndParent,"Destop's DSP Player.\n\n For Playing NGC DSP files\n\n Destop-CZN (c) 2003\n\nv1.6 Modified by hcs","Destop's DSP Player",MB_OK);
}

void init() {
	ReadSettings();
}

void quit() {
}

int play(char *fn) 
{ 
	int maxlatency;
	int thread_id;
	
	//input_file = CreateFile(fn,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,
	//	OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	//if (input_file == INVALID_HANDLE_VALUE) // error opening file
	//{
	//	return 1;
	//}

	if (InitDSPFILE(fn, &dspfile)) return 1;

	fadecount=-1;

	lstrcpy(lastfn,fn);
	paused=0;
	decode_pos_ms=0;
	seek_needed=-1;

	maxlatency = mod.outMod->Open(dspfile.ch[0].header.sample_rate,dspfile.NCH,BPS, -1,-1);
	if (maxlatency < 0) // error opening device
	{
		MessageBox(NULL, "error opening device", "DSP error", MB_OK);
		return 1;
	}
	// dividing by 1000 for the first parameter of setinfo makes it
	// display 'H'... for hundred.. i.e. 14H Kbps.
	mod.SetInfo((dspfile.ch[0].header.sample_rate*BPS*dspfile.NCH)/1000,dspfile.ch[0].header.sample_rate/1000,dspfile.NCH,1);

	// initialize vis stuff
	mod.SAVSAInit(maxlatency,dspfile.ch[0].header.sample_rate);
	mod.VSASetInfo(dspfile.NCH,dspfile.ch[0].header.sample_rate);

	mod.outMod->SetVolume(-666); // set the output plug-ins default volume

	killDecodeThread=0;
	thread_handle = (HANDLE) CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) DecodeThread,(void *) &killDecodeThread,0,&thread_id);
	
	return 0; 
}


void pause() 
{ 
	if (!paused) mod.outMod->Pause(1); 
	paused=1; 
}


void unpause() 
{ 
	if (paused) mod.outMod->Pause(0);
	paused=0;
}


int ispaused() 
{ 
	return paused; 
}


void stop() 
{ 
	if (thread_handle != INVALID_HANDLE_VALUE)
	{
		killDecodeThread=1;
		if (WaitForSingleObject(thread_handle,INFINITE) == WAIT_TIMEOUT)
		{
			MessageBox(mod.hMainWindow,"error asking thread to die!\n","error killing decode thread",0);
			TerminateThread(thread_handle,0);
		}
		CloseHandle(thread_handle);
		thread_handle = INVALID_HANDLE_VALUE;
	}

	CloseDSPFILE(&dspfile);
	
	mod.outMod->Close();
	mod.SAVSADeInit();
}


int getlength() 
{ 
	return dspfile.nrsamples/dspfile.ch[0].header.sample_rate*1000;
}


int getoutputtime() 
{ 
	return mod.outMod->GetOutputTime();
}


void setoutputtime(int time_in_ms) 
{ 
	seek_needed=time_in_ms; 
}


void setvolume(int volume) 
{ 
	mod.outMod->SetVolume(volume); 
}


void setpan(int pan) 
{ 
	mod.outMod->SetPan(pan); 
}


int infoDlg(char *fn, HWND hwnd)
{
	DSPFILE iddsp;
	char msgbuf[512],loopstr[50]="no loop",type[50];
	
	if (!InitDSPFILE(fn,&iddsp)) {

		if (iddsp.ch[0].header.loop_flag) 
			sprintf(loopstr,"loop start: 0x%04x\nloop end: 0x%04x",iddsp.ch[0].header.sa,iddsp.ch[0].header.ea);

		switch (iddsp.ch[0].type) {
		case type_std:
			strcpy(type,"devkit or unrecognized");
			break;
		case type_sfass:
			strcpy(type,"Star Fox Assault/Donkey Konga (Cstr)");
			break;
		case type_mp2:
			strcpy(type,"Metroid Prime 2 (RS03)");
			break;
		case type_pm2:
			strcpy(type,"Paper Mario 2 (STM)");
			break;
		case type_halp:
			strcpy(type,"Super Smash Bros. Melee (HALPST)");
			break;
		}

		sprintf(msgbuf,"%d Hz %s DSP\n%s header type\n%d samples\n%s",iddsp.ch[0].header.sample_rate,
			(iddsp.NCH==2?"stereo":"mono"),type,
			iddsp.nrsamples,loopstr
			);

		MessageBox(mod.hMainWindow,msgbuf,"DSP Info",MB_OK);
		CloseDSPFILE(&iddsp);
	}

	return 0;
}

void getfileinfo(char *filename, char *title, int *length_in_ms)
{
	DSPFILE fidsp;
	
	if (!filename || !*filename)  // currently playing file
	{
		if (length_in_ms) *length_in_ms=getlength();
		if (title) 
		{
			char *p=lastfn+lstrlen(lastfn);
			while (*p != '\\' && p >= lastfn) p--;
			lstrcpy(title,++p);
		}
	}
	else // some other file
	{
		if (length_in_ms) 
		{
			*length_in_ms=-1000;
			if (!InitDSPFILE(filename,&fidsp))
			{
				// these are only second-accurate, but how accurate does this need to be anyway?
				if (!fidsp.ch[0].header.loop_flag) *length_in_ms = fidsp.nrsamples/fidsp.ch[0].header.sample_rate*1000;
				else *length_in_ms=((fidsp.ch[0].header.sa+looptimes*(fidsp.ch[0].header.ea-fidsp.ch[0].header.sa))*14/8/2/fidsp.ch[0].header.sample_rate+masterfadelength)*1000;
			}
			CloseDSPFILE(&fidsp);
		}
		if (title) 
		{
			char *p=filename+lstrlen(filename);
			while (*p != '\\' && p >= filename) p--;
			lstrcpy(title,++p);
		}
	}
}


void eq_set(int on, char data[10], int preamp) 
{ 
	// most plug-ins can't even do an EQ anyhow.. I'm working on writing
	// a generic PCM EQ, but it looks like it'll be a little too CPU 
	// consuming to be useful :)
}

int isourfile(char *fn) { return 0; } 
// used for detecting URL streams.. unused here. strncmp(fn,"http://",7) to detect HTTP streams, etc

// render 576 samples into buf. 
// note that if you adjust the size of sample_buffer, for say, 1024
// sample blocks, it will still work, but some of the visualization 
// might not look as good as it could. Stick with 576 sample blocks
// if you can, and have an additional auxiliary (overflow) buffer if 
// necessary.. 

int get_576_samples(short *buf)
{
	int i;

	for (i=0;i<576;i++) {
		if (looptimes && decode_pos_ms*dspfile.ch[0].header.sample_rate/1000+i >= dspfile.nrsamples) return i*dspfile.NCH*(BPS/8);
		if (dspfile.ch[0].readloc == dspfile.ch[0].writeloc) {
			fillbuffers(&dspfile);
			if (dspfile.ch[0].readloc != dspfile.ch[0].writeloc) return i*dspfile.NCH*(BPS/8);
		}
		buf[i*dspfile.NCH]=dspfile.ch[0].chanbuf[dspfile.ch[0].readloc++];
		if (dspfile.NCH==2) buf[i*dspfile.NCH+1]=dspfile.ch[1].chanbuf[dspfile.ch[1].readloc++];
		if (dspfile.ch[0].readloc>=0x8000/8*14) dspfile.ch[0].readloc=0;
		if (dspfile.ch[1].readloc>=0x8000/8*14) dspfile.ch[1].readloc=0;

		// fade
		if (looptimes && dspfile.ch[0].header.loop_flag && dspfile.nrsamples*1000.0/dspfile.ch[0].header.sample_rate-decode_pos_ms < masterfadelength*1000) {
			buf[i*dspfile.NCH]*=(double)(dspfile.nrsamples*1000.0/dspfile.ch[0].header.sample_rate-decode_pos_ms)/masterfadelength/1000.0;
			if (dspfile.NCH==2) buf[i*dspfile.NCH+1]*=(double)(dspfile.nrsamples*1000.0/dspfile.ch[1].header.sample_rate-decode_pos_ms)/masterfadelength/1000.0;
		}
	}
	return 576*dspfile.NCH*(BPS/8);
}


DWORD WINAPI __stdcall DecodeThread(void *b)
{
	int done=0;
	while (! *((int *)b) ) 
	{
		if (seek_needed != -1)
		{

			if (decode_pos_ms>seek_needed) {

					InitDSPFILE(NULL, &dspfile);

					fadecount=-1;
	
					decode_pos_ms=-1; // -1 ms, force it to work even if we want to seek to 0
			}
	
			while (decode_pos_ms<seek_needed) {
				get_576_samples((short*)sample_buffer);
				mod.outMod->Flush(decode_pos_ms);
				decode_pos_ms+=(576.0*1000.0)/(double)dspfile.ch[0].header.sample_rate;
			}
			seek_needed=-1;
		}
		if (done)
		{
			mod.outMod->CanWrite();
			if (!mod.outMod->IsPlaying())
			{
				PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
				return 0;
			}
			Sleep(10);
		}
		else if (mod.outMod->CanWrite() >= ((576*dspfile.NCH*(BPS/8))<<(mod.dsp_isactive()?1:0)))
		{	
			int l=576*dspfile.NCH*(BPS/8);
			l=get_576_samples((short*)sample_buffer);
			if (!l) 
			{
				done=1;
			}
			else
			{
				if (l==576*dspfile.NCH*(BPS/8)) {
					mod.SAAddPCMData((char *)sample_buffer,dspfile.NCH,BPS,mod.outMod->GetWrittenTime());
					mod.VSAAddPCMData((char *)sample_buffer,dspfile.NCH,BPS,mod.outMod->GetWrittenTime());
				}
				decode_pos_ms+=(576.0*1000.0)/(double)dspfile.ch[0].header.sample_rate;
				if (mod.dsp_isactive()) l=mod.dsp_dosamples((short *)sample_buffer,l/dspfile.NCH/(BPS/8),BPS,dspfile.NCH,dspfile.ch[0].header.sample_rate)*(dspfile.NCH*(BPS/8));
				mod.outMod->Write(sample_buffer,l);
			}
		}
		else Sleep(20);
	}
	return 0;
}


In_Module mod = 
{
	IN_VER,
	"Destop's DSP Player v1.6 improved version 11/12/05"
#ifdef _DEBUG
	" (DEBUG)"
#endif
	,
	0,	// hMainWindow
	0,  // hDllInstance
	"DSP\0DSP Audio File (*.DSP)\0GCM\0DSP Audio File (*.GCM)\0HPS\0HALPST Audio File (*.HPS)\0"
	,
	1,	// is_seekable
	1, // uses output
	config,
	about,
	init,
	quit,
	getfileinfo,
	infoDlg,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,
	
	getlength,
	getoutputtime,
	setoutputtime,

	setvolume,
	setpan,

	0,0,0,0,0,0,0,0,0, // vis stuff


	0,0, // dsp

	eq_set,

	NULL,		// setinfo

	0 // out_mod

};

__declspec( dllexport ) In_Module * winampGetInModule2()
{
	return &mod;
}

int main (void)
{
	return 0;
}
